package it.polimi.dei.esercitazioniIngegneriaDelSoftware.swing.calculator;

import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

/**
 * Rappresenta una Interfaccia utente grafica per 
 * una calcolatrice.
 * @author andrea
 *
 */
public class CalculatorGUI implements CalculatorUI {

  // finestra principale
  private final JFrame calculatorFrame;
  
  // rappresenta il risultato
  private JLabel result;
  // rappresenta l'operando corrente
  private JTextField value;

  // lista di bottoni per le cifre
  private List<JButton> digitsButtons;
  // lista di bottoni per le operazioni
  private List<JButton> operationButtons;
  // altri bottoni: virgola, annulla, =
  private JButton dotButton, cancelButton, equalsButton;

  // modello, usato SOLO negli handler (ActionListener)
  private CalculatorModel calculatorModel;

  /**
   * Crea una nuova GUI.
   * @param math Un modello di calcolatrice.
   */
  public CalculatorGUI(CalculatorModel math) {
    this.calculatorFrame = new JFrame("Simple Calculator");
    this.calculatorModel = math;
    setup();
  }
  
  // ==============================================================
  // Questa classe ha tre parti:
  // 1. Metodi privati per la costruzione della gui,
  // 2. Metodi pubblici chiamati dal modello per aggiornare la gui
  // 3. Classi interne che rappresentano gli handler per gestire
  //    gli eventi della GUI.
  // ==============================================================
  
  // ==============================================================
  // METODI PRIVATI PER LA COSTRUZIONE DELLA GUI
  // ==============================================================

  // Definisce la struttura del panel del frame principale
  // e ne costruisce le sue parti
  private void setup() {
    setupButtons();

    this.calculatorFrame.setLayout(new BorderLayout());

    JPanel upperPanel = buildUpperPanel();
    this.calculatorFrame.add(upperPanel, BorderLayout.NORTH);

    JPanel centralPanel = buildCentralPanel();
    this.calculatorFrame.add(centralPanel, BorderLayout.CENTER);

    JPanel operationsPanel = buildOperationsPanel();
    this.calculatorFrame.add(operationsPanel, BorderLayout.EAST);

    this.calculatorFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    this.calculatorFrame.pack();
    this.calculatorFrame.setVisible(true);
    // lo rendo non resizable per evitare dimensioni in cui i componenti
    // non sono piu' visibili.
    this.calculatorFrame.setResizable(false);
  }

  // costruisce il pannello sulle operazioni
  private JPanel buildOperationsPanel() {
    JPanel operationsPanel = new JPanel(new GridLayout(5, 1));

    for (JButton operationButton: this.operationButtons) {
      operationsPanel.add(operationButton);
    }

    operationsPanel.add(equalsButton);
    return operationsPanel;
  }

  // costruisce il pannello centrale con le operazioni
  private JPanel buildCentralPanel() {
    JPanel centralPanel = new JPanel(new GridLayout(4, 3));
    for (JButton digitButton: this.digitsButtons) {
      centralPanel.add(digitButton);
    }

    centralPanel.add(dotButton);
    centralPanel.add(this.cancelButton);
    return centralPanel;
  }

  // costruisce il pannello superiore con label del risulato
  // e textfield dell'operando
  private JPanel buildUpperPanel() {
    JPanel upperPanel = new JPanel(new GridLayout(2, 1));
    result = new JLabel("0.0");
    upperPanel.add(result);
    value = new JTextField("");

    // Non deve essere editabile direttamente dall'utente
    value.setEditable(false); 
    upperPanel.add(value);
    return upperPanel;
  }

  private void setupButtons() {
    // crea i bottoni per le cifre
    createDigitButtons();
    // crea i bottoni per le operazioni
    createOperationButtons();

    this.cancelButton = new JButton("C");
    this.cancelButton.addActionListener(new CancelButtonHandler());
  }

  
  
  // crea i bottoni corrispondenti alle cifre e alla virgola
  private void createDigitButtons() {
    this.digitsButtons = new ArrayList<>(10);

    for (int i = 1; i <= 10; i++) {
      int digit = i % 10;
      JButton digitButton = new JButton(String.valueOf(digit));
      DigitsButtonHandler digitButtonHandler = new DigitsButtonHandler(digit);
      digitButton.addActionListener(digitButtonHandler);
      this.digitsButtons.add(digitButton);
    }

    this.dotButton = new JButton(".");
    this.dotButton.addActionListener(new DotButtonHandler());
  }

  // crea i bottoni corrispondenti alle operazioni, compresa l'uguaglianza.
  private void createOperationButtons() {
    this.operationButtons = new ArrayList<>(Operation.values().length);

    for (Operation operation: Operation.values()) {
      /* non usare -name()- per le enum, ma implementare la toString, come in questo caso */
      JButton operationButton = new JButton(operation.toString());
      OperationButtonHandler operationButtonHandler = new OperationButtonHandler(operation);
      operationButton.addActionListener(operationButtonHandler);
      this.operationButtons.add(operationButton);
    }		

    this.equalsButton = new JButton("=");
    this.equalsButton.addActionListener(new EqualsButtonHandler());
  }
  
  // ==============================================================
  //  metodi chiamati dal modello per aggiornare la vista
  // ==============================================================
  
  @Override
  public void setValue(double accumulatedValue) {
    this.result.setText(String.valueOf(accumulatedValue));  
    this.value.setText("");
  }

  @Override
  public void setDigits(boolean enabled) {
    for (JButton digit: this.digitsButtons) {
      digit.setEnabled(enabled);
    }
    this.dotButton.setEnabled(enabled);
  }

  @Override
  public void setEquals(boolean enabled) {
    this.equalsButton.setEnabled(enabled);
  }

  // ==============================================================
  // HANDLER per la gestione degli eventi della GUI
  // Essendo le componenti tutte bottoni, usiamo solo 
  // ActionListener. Il codice dei metodi dell'handler e'
  // in generale breve, e contiene chiamate corrispondenti
  // ai metodi del modello e raramente ai metodi della vista.
  // Quando vedremo i design pattern, vedremo come strutturare
  // questo codice in maniera piu' pulita ed elegante.
  // E' anche possibile non usare classi interne e usare classi
  // separate.
  // ==============================================================
  
  /**
   * Handler per la gestione di un click di un bottone
   * che rappresenta una operazione. Agisce sul modello 
   * settando l'operazione corrente.
   */
  class OperationButtonHandler implements ActionListener {

    private final Operation operation;

    public OperationButtonHandler(Operation operation) {
      this.operation = operation;
    }

    public void actionPerformed(ActionEvent e) {
      calculatorModel.setCurrentOperation(operation);
    }

  }
  
  /**
   * Handler per la gestione di un click di un bottone
   * che rappresenta una cifra. Agisce sul modello 
   * settando l'operando corrente.
   */
  class DigitsButtonHandler implements ActionListener {

    private final int digit;

    DigitsButtonHandler(int digit) {
      this.digit = digit;
    }

    public void actionPerformed(ActionEvent e) {
      Double val = Double.valueOf(value.getText() + digit);
      value.setText(value.getText() + digit);
      calculatorModel.setOperand(val);
    }

  }

  /**
   * Handler per la gestione di un click del bottone
   * che rappresenta il tasto uguale (=). Agisce sul modello 
   * calcolando un nuovo risultato.
   */
  class EqualsButtonHandler implements ActionListener {

    @Override
    public void actionPerformed(ActionEvent e) {
      calculatorModel.compute();
    }
  }

  /**
   * Handler per la gestione di un click del bottone
   * che rappresenta la virgola. Agisce solo sulla vista
   * aggiungendo il simbolo corrispondente nella JTextField
   * che contiene l'operando corrente.
   */
  class DotButtonHandler implements ActionListener {

    @Override
    public void actionPerformed(ActionEvent arg0) {
      if (!value.getText().contains("."))
        value.setText(value.getText() + ".");
    }
  }

  /**
   * Handler per la gestione di un click del bottone
   * che resetta lo stato della calcolatrice. Agisce sul modello 
   * in maniera corrispondente.
   */  
  class CancelButtonHandler implements ActionListener {

    @Override
    public void actionPerformed(ActionEvent e) {
      calculatorModel.reset();
    }

  }
}
